//      create_metafs.c
//      
//      Copyright 2010 Leszek Kattinger <letesch@gmx.de>
//      
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

#define _XOPEN_SOURCE 500
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <glib.h>
#include <assert.h>
#include <signal.h>
#include <time.h>
#include "metafs.h"
#include <sys/time.h> 

/*
gcc `pkg-config --cflags --libs glib-2.0` metafs.c create_metafs.c -o create_metafs

run parameter: <filename> <file_size> <block_size> [f]
 
example:  2 GB volume size 64 KB block size, format all blocks

./create_metafs data 2000 64 f

*/

static void signalHandler(int signal)
{
  printf("Caught signal %d\n", signal);  
 
	int c;

  if(signal != SIGINT)
      return;
  else {
      printf("\n Quit ? (j/N) : ");
      c=getchar();
      if(c == 'j')				 
         exit (EXIT_FAILURE);
      else
         return;
   }     
}

int main ( int argc, char *argv[] )
{  	
	signal(SIGINT, & signalHandler);
	
	Super * s = (Super *)initSuper();
		
  uint64_t volume_size_mb = g_ascii_strtoull( argv[2], NULL, 10) ;   // convert MB -> Byte
	
	s->volume_size = volume_size_mb * 1024 * 1024; 
	
	s->block_size = (uint32_t)g_ascii_strtoull( argv[3], NULL, 10) * 1024; // convert KB -> Byte
 			
	if( argc < 4)
	{
		printf ("Enter FS Parameter: <file/device> <file_size in MB> <block_size in KB> <f = format all>\n");
		return EXIT_FAILURE;	
	}
	
	calculateSuper(s);		
	
	if(s->bitmap_bsize > s->block_size) {
		fprintf(stderr, "Problem format device - block_size too small for bitmap! \n");
    return false;
	}	
			
	char buff[s->block_size+1];
	memset(buff, 0,  s->block_size);
	
	s->file  = open(argv[1], O_RDWR | O_CREAT, 0777);
	
	if( s->file  == -1 ) {
		fprintf(stderr, "Problem format device - can't open  %s \n", argv[1] );
    return false;	}	
				 
	uint64_t i, max_blocks = 10;  // default - format 10 blocks if not specified 
	if( argc == 5 )
	max_blocks = ( strcmp( argv[4], "f") == 0) ? s->blocks : 10;	
	
	puts("#############################################################################");	
	printf("Start format Filesystem on:  %s  with: %"PRIu64" blocks \n", argv[1] , max_blocks);	 
	printf ("  Block (%"PRIu32" KB) write time [ms]  \n ", s->block_size/1024);	
	
	struct timeval tmp,  tmp_start;
	long long unsigned count_time_usec = 0;	
	long long unsigned count_time_sec = 0;
	long long unsigned tmp_lapsed_usec = 0;
	long long unsigned tmp_lapsed_sec = 0;  
  	
	for( i = 0; i< max_blocks; i++)		 
	{		
			gettimeofday (&tmp_start, NULL);	// start time for block write
			
			ssize_t ret = pwrite(s->file, buff, s->block_size, i * s->block_size); // format => write zeros to block 
			
			if(ret < s->block_size ) {
      fprintf(stderr, "Problem format fs - written Bytes %llu @ block %"PRIu64"\n", (long long unsigned) ret, i );
      return false;
			}				 
			gettimeofday (&tmp, NULL);				// stop time for block write
			
			if (tmp_start.tv_usec > tmp.tv_usec) 
			{  
				tmp.tv_usec += 1000000;  tmp.tv_sec--; 
			} 		
							
			tmp_lapsed_usec += (tmp.tv_usec - tmp_start.tv_usec);
			tmp_lapsed_sec += (tmp.tv_sec - tmp_start.tv_sec);
							
			count_time_usec += tmp_lapsed_usec;
			count_time_sec += tmp_lapsed_sec; 			
								
			if( i%16 == 0 ) printf("\n [%" PRIu64"] ", i);		
					
			printf (" %llu " ,tmp_lapsed_usec/1000 );
			tmp_lapsed_usec = tmp_lapsed_sec = 0;  // reset tmp time	
	}			
	printf ("\n  done! \n ");		
   
	printf (" \n writen: %"PRIu64" blocks in %llu.%llu seconds \n",i , count_time_usec/1000000,  count_time_usec%1000 ) ;
 
  if((count_time_usec/1000000) > 0 )
  printf(" => average block-write speed: %3.2f MB/s \n ",  (double)( volume_size_mb  / (count_time_usec/1000000))   );
    
 //  printf(" => average block-write time: %llu us/block  (%llu ms/block)\n" , count_time_usec/s->blocks,  (count_time_usec/1000)/s->blocks);
 
		
	printf ("----------------------\n Write data structures \n ----------------------\n");	
	
	initBitmap(s);		
	
	createSuper(s);	
	
	writeSuper(s);		//displaySuper(s);	
	
	writeBitmap(s);	  //displayBitmap(s);
	
	printf (" \n done! \n") ;	
	   
	printf ("--------------------- \n Test Filesystem \n");		
		
	readSuper(s);		displaySuper(s);				
		
	readBitmap(s);	//dsplayBitmap(s);
	
	close(s->file);		
	
	printf (" \n done!  \n") ;
	
  return EXIT_SUCCESS;

}
