//      fuse_metafs.c
//      
//      Copyright 2010 Leszek Kattinger <letesch@gmx.de>
//      
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

/*
gcc -Wall -lpthread `pkg-config --cflags --libs glib-2.0 fuse` metafs.c fuse_metafs.c -o fuse_metafs 
run:
./fuse_metafs -d -o default_permissions -o allow_other /tmp/metafs /dev/xxx
*/ 

#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#include <unistd.h>
#undef  _XOPEN_SOURCE
#else
#include <unistd.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <glib.h>
#include <assert.h>
#include <signal.h>

#define FUSE_USE_VERSION 26
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fuse.h>
#include "metafs.h"

static char * device;

static Filesystem filesystem;

static int metafs_getattr(const char *path, struct stat *stbuf)
{		
	Entry * e = mefs_lookupPath(path);
	if(e == NULL)
		return -ENOENT;

	stbuf->st_mode = e->a.st_mode;
	stbuf->st_atime = e->a.st_atime;
	stbuf->st_mtime = e->a.st_mtime;
	stbuf->st_ctime = e->a.st_ctime;	
	stbuf->st_uid = e->a.st_uid;
	stbuf->st_gid = e->a.st_gid;
	stbuf->st_nlink = 1;
	stbuf->st_size = 0;
	
	return 0;
}

struct readdir_callback_data
{
	fuse_fill_dir_t filler;
	void *buf;	
};


gboolean niter_all_entries(gpointer name, Entry * e, void *data )
{
  //displayEntry(e);
  
  struct readdir_callback_data *de = data;  
  
  de->filler(de->buf, e->name, NULL, 0);	  
  
  return FALSE;
}


static int metafs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
		       off_t offset, struct fuse_file_info *fi)
{	
	(void) offset;
	(void) fi;		
	//printf("<metafs_readdir > path = %s \n", path);
	
	Entry * e = mefs_lookupPath(path);		// find dir entry, load dir struct
	if(e == NULL )
		return -ENOENT;		
	//printf("<metafs_readdir > e->name = %s e->data =  %"PRIu64"\n", e->name, e->data);
	
	Dir * d = mefs_lookupDir(e->data);		// get pointer to dir struct in RAM
	if(d == NULL )
	{
	 d = mefs_loadDir(e->data);
	printf("<metafs_readdir > d->dir_nr = %"PRIu64"  \n", d->dir_nr);	
	}
	
  // output in alphabetic order 
  
	struct readdir_callback_data de;
						
	de.filler = filler;	 
  de.buf = buf;				  
	g_tree_foreach(d->t, (GTraverseFunc)niter_all_entries,  &de );
  
	
  // unsorted output 
  
  /* struct stat st;  
				 
	Block * b = d->b;
	while( b != NULL )
	{
		e = b->e;		
		while( e != NULL)
		{
			//struct stat st;
			//memset(&st, 0, sizeof(st));
			//filler(buf, e->name, &st, 0);	
			filler(buf, e->name, NULL, 0);	
			e = e->next;
		}
		b = b->next;
	}
	*/
	return 0;
}

static int metafs_mkdir(const char *path, mode_t mode)
{
	mode =  S_IFDIR | 0777 ;
	
	char * dirname = g_path_get_dirname(path);
	
	Entry * e = mefs_lookupPath(dirname);
	if(e == NULL )
		return -ENOENT;		
		
	Dir * d = (Dir *)mefs_lookupDir(e->data);
	if(d == NULL )
		d = (Dir *)mefs_loadDir(e->data);
	else if(d == NULL )
		return -EIO;	
		
	char * basename = g_path_get_basename(path);	
	
	e = (Entry *)mefs_initEntry(basename, mode);
	if(e == NULL)
		return -ENOMEM;
	
	if( !mefs_create(filesystem.s, d, e) )
		return -EIO;
		
	return 0;
}


static int metafs_mknod(const char *path, mode_t mode, dev_t rdev)
{	
	mode =  S_IFREG | 0777  ; 
		
	char * dirname = g_path_get_dirname(path);
	
	Entry * e = mefs_lookupPath(dirname);
	if(e == NULL )
		return -ENOENT;		
		
	Dir * d = (Dir *)mefs_lookupDir(e->data);
	if(d == NULL )
		d = (Dir *)mefs_loadDir(e->data);
	else if(d == NULL )
		return -EIO;	
	
	char * basename = g_path_get_basename(path);	
	
	e = (Entry *)mefs_initEntry(basename, mode);
	
	if( !mefs_create(filesystem.s, d, e) )
		return -EIO;	
	
	return 0;
}

static int metafs_create(const char *path, mode_t mode)
{	
	//mode =  S_IFREG | 0777 ; 	
		
	char * dirname = g_path_get_dirname(path);	
	
	Entry * e = mefs_lookupPath(dirname);
	if(e == NULL )
		return -ENOENT;		
		
	Dir * d = (Dir *)mefs_lookupDir(e->data);
	if(d == NULL )
		d = (Dir *)mefs_loadDir(e->data);
	else if(d == NULL )
		return -EIO;	
	
	
	char * basename = g_path_get_basename(path);	
	
	e = (Entry *)mefs_initEntry(basename, mode);		
	
	if( !mefs_create(filesystem.s, d, e) )
		{return -EIO;	}
		
	return 0;
	
}

static int metafs_rmdir(const char *path)
{	
	Entry * e = mefs_lookupPath(path);
	if(e == NULL  )
		return -ENOENT;		
	
	if ( !removeDir(filesystem.s, e->data) )
		return -ENOTEMPTY;		
		
	if( !mefs_delete(filesystem.s, e) )
			return -EIO;
			
	return 0;
}

static int metafs_unlink(const char *path)
{
	Entry * e = mefs_lookupPath(path);
	if(e == NULL )
		return -ENOENT;			
	
	if( !mefs_delete(filesystem.s, e) )
		return -EIO;		

	return 0;
}

static int metafs_rename(const char *from, const char *to)
{	
	
		Entry * efrom = mefs_lookupPath(from);    
		if(efrom == NULL  )
		return -ENOENT;		
			
		char * dirname = g_path_get_dirname(to);	
	
		Entry * e = mefs_lookupPath(dirname);
		if(e == NULL )
		return -ENOENT;		
		
		Dir * dto = (Dir *)mefs_lookupDir(e->data);
		if(dto == NULL )
		dto = (Dir *)mefs_loadDir(e->data);
			else if(dto == NULL )
			return -EIO;	
		
		char * basename = g_path_get_basename(to);				
				
	  Entry * eto = (Entry *)mefs_initEntry(basename, efrom->a.st_mode);	
	  
	  if( !mefs_create(filesystem.s, dto, eto) )
			{return -EIO;	}			
		
		eto->data = efrom->data;				
	  
	  if( !mefs_rename(filesystem.s, efrom) )
		return -EIO;				
     	
	return 0;
}

static int metafs_symlink(const char *from, const char *to)
{
	
	mode_t mode =  S_IFLNK | 0775 ; 
		
	char * dirname = g_path_get_dirname(to);
	
	Entry * e = mefs_lookupPath(dirname);
	if(e == NULL )
		return -ENOENT;		
		
	Dir * d = (Dir *)mefs_lookupDir(e->data);
	if(d == NULL )
		d = (Dir *)mefs_loadDir(e->data);
	else if(d == NULL )
		return -EIO;	
	
	char * basename = g_path_get_basename(to);	
	
	e = (Entry *)mefs_initEntry(basename, mode);	
	
	char * n = 	strcat(e->name, "hup");
	
	e->name_len = strlen(n);

	// tricky - left unimplementet for idea / decision TODO
    
	//gchar * l = "\0hal";
	
	//e->name_len = strlen(basename) + 4;
	
	//e->name = g_string_append(e->name, l  );	
	
	if( !mefs_create(filesystem.s, d, e) )
		return -EIO;	
	
	return 0;
}

static int 	metafs_readlink (const char *path, char *buf, size_t size)
{	
  // tricky - left unimplementet for idea / decision TODO
  
	Entry * e = mefs_lookupPath(path);
	
	buf = "/hallo";
	
	size = 6;
	
	return 0;
}


static int metafs_chmod(const char *path, mode_t mode)
{	
	Entry * e = mefs_lookupPath(path);
	if(e == NULL)
		return -ENOENT;	
		
	if( !mefs_chmod(e, mode))
		return -EIO;		
	
	return 0;
}

static int metafs_chown(const char *path, uid_t uid, gid_t gid)
{	
	Entry * e = mefs_lookupPath(path);
	if(e == NULL)
		return -ENOENT;	
		
	if( !mefs_chown(e, uid, gid))
			return -EIO;				
	
	return 0;
}

static int metafs_utimens(const char *path, const struct timespec ts[2])
{				
	Entry * e = mefs_lookupPath(path);
	if(e == NULL)
		return -ENOENT;	
		
	struct timeval tv[2];

	tv[0].tv_sec = ts[0].tv_sec;
	tv[0].tv_usec = ts[0].tv_nsec / 1000;
	tv[1].tv_sec = ts[1].tv_sec;
	tv[1].tv_usec = ts[1].tv_nsec / 1000;
	
	if ( !mefs_utimens(e, tv) )
		return -EIO;
	
	return 0;
}

static int metafs_access(const char *path, int mask)
{
	Entry * e = mefs_lookupPath(path);
	if(e == NULL)
		return -ENOENT;	
	return 0;
}

static int metafs_open(const char *path, struct fuse_file_info *fi)
{
	Entry * e = mefs_lookupPath(path);
	if(e == NULL)
		return -ENOENT;	
	return 0;
}

static int metafs_read(const char *path, char *buf, size_t size, off_t offset,
		    struct fuse_file_info *fi)
{
	Entry * e = mefs_lookupPath(path);
	if(e == NULL)
		return -ENOENT;	
	return size;
}

static int metafs_write(const char *path, const char *buf, size_t size,
		     off_t offset, struct fuse_file_info *fi)
{
	Entry * e = mefs_lookupPath(path);
	if(e == NULL)
		return -ENOENT;	
	return size;
}


static int metafs_statfs(const char *path, struct statvfs *stbuf)
{
	/*
	int res;
	res = statvfs(path, stbuf);
	if (res == -1)
		return -errno;
*/	
  // TODO find out what var names are used for status ...
  	
	stbuf->f_namemax = (unsigned long) 255;
	//stbuf->f_bfree = countFreeBlocks(filesystem.s);
	return 0;
}

static int metafs_release(const char *path, struct fuse_file_info *fi)
{
 	// Just a stub.	 This method is optional and can safely be left
	//   unimplemented 
	(void) path;
	(void) fi;
	return 0;
}

static int metafs_fsync(const char *path, int isdatasync,
		     struct fuse_file_info *fi)
{
	/* Just a stub.	 This method is optional and can safely be left
	   unimplemented */
	/*
	(void) path;
	(void) isdatasync;
	(void) fi;
	*/
	return 0;
}

static void * metafs_init( struct fuse_conn_info *conn)
{	
  // extend mount parameter for static set log to data TODO 	

	filesystem = (Filesystem)mefs_loadFilesystem( device, "data");	
	return &filesystem; 
}

static void metafs_destroy(void *userdata)
{
	mefs_destroy();
}

static struct fuse_operations metafs_oper = {
  
	.getattr	= metafs_getattr,	
	.readdir	= metafs_readdir,
  
	.mknod		= metafs_mknod,   // mknode vs. create -> ?
	.create		= metafs_create,	
	.mkdir		= metafs_mkdir,	
	.unlink		= metafs_unlink,
	.rmdir		= metafs_rmdir,
	.rename		= metafs_rename,	
  
	.chmod		= metafs_chmod,
	.chown		= metafs_chown,	
	.utimens	= metafs_utimens,
  
	.init 		= metafs_init,		
	.destroy  = metafs_destroy,
  			
	.readlink	= metafs_readlink,
	//.link			= metafs_link,
	.symlink	= metafs_symlink,

	.access		= metafs_access, 
	.open		= metafs_open,
	.read		= metafs_read,
  
	.write		= metafs_write,
	.statfs		= metafs_statfs,
	//.truncate	= metafs_truncate,
	//.release	= metafs_release,
	//.fsync		= metafs_fsync,
};

int main(int argc, char *argv[])
{		
	//umask(0);	// ???	
	
	printf("\a");
	
	if( argc < 3)		// min fs parameter <mountpoint> <device>
	{
		printf ("Enter at least 2 FS parameter: <mountpoint> <device / file> \n");
		exit (EXIT_FAILURE);
	}
	// parameter exchange TODO - works but there is better way + add log/data location selection
  
	size_t len = strlen(argv[argc-1])+1;
	
	device = (char *)calloc(1, len);
	
	strcpy(device, argv[argc-1]);
	//printf("args = %d  storage device = %s \n",argc, device);
	
	char * newargv[argc-1];		
	
	int i;
	for(i=0 ; i<argc-1; i++){	
		newargv[i] = argv[i];
		//printf("arg %i = %s \n", i, argv[i]);
	}	
	newargv[argc-1] = argv[argc];
	//printf("newargv[argc-1]= %s  argv[argc]; = %s \n",newargv[argc-2], argv[argc-1] );
	--argc;
	
	return fuse_main(argc, newargv, &metafs_oper, NULL);
}



