//      metafs.c
//      
//      Copyright 2010 Leszek Kattinger <letesch@gmx.de>
//      
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

#define _XOPEN_SOURCE 500
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <grp.h>
#include <time.h>
#include <locale.h>
#include <langinfo.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <glib.h>
#include <pthread.h>
#include <assert.h>
#include <signal.h>
#include "metafs.h" 

// time to sync file system metadata:
#define WRITE_BACK_TIME 5

#define WRITE_LOCK  pthread_mutex_lock(& fs.write_mutex);
#define WRITE_UNLOCK  pthread_mutex_unlock(& fs.write_mutex);

static Filesystem fs;

static bool writeBackBlock(Block * b, int file, off_t offset, bool write_full_block);

#define dbg  1

/**
 * Signal handler for catching signals
 */
static void signalHandler(int signal)
{
   printf("Caught signal %d\n", signal);   
   mefs_destroy();  
}

static void enforce_write_back()
{
  WRITE_LOCK  
  pthread_cond_signal(& fs.enforce_write_back_cond);  
  WRITE_UNLOCK
}

/** 
 * TODO - remove blocks from dirty / writeback list 
 */ 
static void remove_from_dirty_block_list(Block * b)
{
	if(dbg > 0)  printf("<remove_from_dirty_block_list>  %"PRIu64" \n", b->block_nr);	
	WRITE_LOCK
	
	GList * listElem =  g_list_find(fs.dirty_block_list, (gconstpointer)b);   
	
	fs.dirty_block_list = g_list_delete_link(fs.dirty_block_list, listElem);
	
	WRITE_UNLOCK
}


/**
 *  write_back function for log and meta data blocks
 *  TODO - settings for log (mount parameter) are involved  - log is not used now
 */ 

static void * write_back_thread_function(void *vptr_args)
{
  while(true)
  {    
    struct timespec timeout;
    struct timeval now;
    //printf("Syncing now\n");
    WRITE_LOCK
    
    gettimeofday(&now, NULL);
    timeout.tv_sec = now.tv_sec + WRITE_BACK_TIME;
    timeout.tv_nsec = now.tv_usec * 1000;
    
    // http://sourceware.org/pthreads-win32/manual/pthread_cond_init.html
    int ret = pthread_cond_timedwait(& fs.enforce_write_back_cond, & fs.write_mutex, & timeout); 
    
    Block * b = NULL; 
  
    
    // phase 1 - write log to backing store
    // iterate / walk through list
    /*
    GList * curElem = g_list_first (fs.dirty_block_list);
    
    uint32_t listElements = g_list_length(fs.dirty_block_list);
    
    off_t offset = 0;

    pwrite( fs.s->data_file, & listElements, sizeof(uint32_t), offset );
      
    offset += sizeof(uint32_t);
    
    if(curElem == NULL){
      WRITE_UNLOCK
      continue;
    }
        
    while( curElem != NULL)
    {      
      b = ((Block* ) curElem->data);	// write block number:
      
      pwrite( fs.s->data_file, & b->block_nr, sizeof(uint64_t), offset );
      
      offset += sizeof(uint64_t);
      
      writeBackBlock(b, fs.s->data_file, offset, false);
      
      offset += b->bs->fillpos;
      
      curElem = curElem->next;
      
      printf("Writing to LOG: block number %"PRIu64" \n", b->block_nr);
     
    }  
    */     
     //displayDirtyList();
    // phase 2 - sync MD to metadata file
     				
    
		while(g_list_first (fs.dirty_block_list) != NULL)		
		{			
			
			GList * listElem = g_list_first(fs.dirty_block_list);	
					
			b = listElem->data;	// FIFO			
			
			printf("Writing to DEVICE: block number %"PRIu64" \n", b->block_nr);
			
			off_t offset =  fs.s->block_size * b->block_nr;
			
			writeBackBlock(b, fs.s->file, offset, true);		// write performance tuning ture = full_block
			
			b->dirty = 0;
			
			fs.dirty_block_list = g_list_delete_link(fs.dirty_block_list, listElem);					
		}		
		
		//if(bitmap_update) writeBitmap(fs.s);	// only when dirty blocks written back
		//writeSuper(fs.s);
    //printf("\n done!\n");
		
    if(fs.closed){
      WRITE_UNLOCK
      return NULL;
    }
    WRITE_UNLOCK        
   }
}

bool displayDirtyList()
{
	puts("<displayDirtyList>------------------");
	g_list_foreach(fs.dirty_block_list, (GFunc)displayBlockStat, NULL);
	puts("<-----------------------------------");
	return true;
}

static inline void setBlockDirty( Block * b)
{ 	
	if(b->dirty == 0)
	{
	   b->dirty = 1;
	   fs.dirty_block_list = g_list_append( fs.dirty_block_list, b);
	   fs.dirty_block_count += 1;
	   //displayDirtyList();	   
	}
}

gboolean iter_all_entries(gpointer name, Entry * e)
{
  displayEntry(e);
  return FALSE;
}

void displayDir(Dir * d)
{		
	printf("--<displayDir>  Tree  dir_nr: %"PRIu64" - %p  nNodes %d\n",
																d->dir_nr, d, g_tree_nnodes(d->t));
															
	g_tree_foreach(d->t, (GTraverseFunc)iter_all_entries, NULL);
}


gboolean iter_all_dirs( gpointer nr, Dir * d)
{
  displayDir(d);	
  return FALSE;
}

bool displayEntry(Entry * e)
{
	if( e == NULL ) 
		return false;	
	 
	printf("<displayEntry>[%s] data [%"PRIu64"] mode [%"PRIu32"] \n",	e->name,	e->data,  e->a.st_mode );
	//	displayBlockStat(e->block_ptr);
	
	return true;							
}
	
void displayLoadedDirs()
{
	puts("------------------------------------");
	printf("<displayLoadedDirs>  Tree nNodes %d \n", g_tree_nnodes(fs.dt) );
	puts("------------------------------------");
	g_tree_foreach(fs.dt, (GTraverseFunc)iter_all_dirs, NULL);
}

bool displayBlockEntries(Block * b)
{	
	Entry * e = b->e;
	
	while( e != NULL)
	{
		displayEntry(e);
		e = e->next;
	}
	return true;
}

bool displayBlockStat(Block * b)
{
	if( b == NULL) {
	fprintf(stderr, "Problem display_blockstat - no block! \n");
	return false;}
		
	printf("<displayBlockStat> block %"PRIu64" dirty %"PRIu8" fillpos %"PRIu32" ext_block_nr %"PRIu64" \n",
															b->block_nr, b->dirty, b->bs->fillpos, b->bs->ext_block_nr);
					
	return true;
}

/**
 * Initialize an entry with name and attributes 
 *
 * Call before create a new dir or file entry
 *
 * @param name containing a file or directory name.
 * @param mode uint32_t containing the entry mode bits
 * @return e pointer to a new entry
  */
Entry * mefs_initEntry(char * name, mode_t mode)
{
	Entry * e = (Entry *) calloc(1, sizeof(Entry));	
	assert( e );	
	
	e->name = name;
		
	e->name_len = (uint8_t) strlen(e->name);
	
	e->a.st_mode = mode;	
			
	uint64_t curtime = time(NULL);		
	e->a.st_ctime = curtime;		
	e->a.st_atime = curtime;
	e->a.st_mtime = curtime;	
	e->a.st_uid = getuid();
	e->a.st_gid = getgid();	
	
	//fs.s->fillpos += e->name_len + CONST_PHYSICAL_ENTRY_SIZE;		// count new entry to cached bytes
				
	printf("<mefs_initEntry>"); displayEntry(e);
								
	return e;
}

Dir * createDir_noLock(uint64_t block_nr)
{
	Dir * d = (Dir *) initDir( block_nr );
			
	Block * b = (Block *) initBlock( block_nr );
		
	appendBlock(d,b);	
		
	appendDir(d);	
	
	setBlockDirty(b);		
	
	puts("<createDir>");	
					
	return d;
}

Block * findFreeBlock_noLock(Super * s, Dir * d, Entry * e)	
{		
	if(d == NULL || e == NULL) {
		fprintf(stderr, "Problem findFreeBlock");
		return false;
	}	
	
	Block * b = d->b;			// find block to append entry
		
	bool found = false;
		
	while(b != NULL) 			// search block with enough space for new entry
	{
		if(b->bs->fillpos + e->name_len + CONST_PHYSICAL_ENTRY_SIZE < s->block_size - CONST_BLOCK_HEADER_SIZE )
		{ 
			found = true;  break; 
		}
		b = b->next;	
	}		
	// if no spce in allocated blocks -> get new ext_block_nr
	if( found == false)	
	{ 
		uint64_t ext_block_nr = getFreeBlock_noLock(s);	
						
		d->last_b->bs->ext_block_nr = ext_block_nr;		// set ext_block_nr		
		
		setBlockDirty(d->last_b);		// set dirty => ext_block_nr changed
						
		b = (Block * ) initBlock( ext_block_nr );			
				
		if( !appendBlock(d,b) )	{
			fprintf(stderr, "%s \n", "error append_node" );
		}
		setBlockDirty(d->last_b);		// set new block dirty => overwrite old val
	}		
	printf(" <findFreeBlock> "); //displayBlockStat(b);
	return b;
}
	
bool mefs_create(Super * s, Dir * d, Entry * e)
{	
	puts("<create>");	
	if(d == NULL || e == NULL)	{
		fprintf(stderr, "Problem create");
		return false;
	}		
	WRITE_LOCK	
	
	Block * b = (Block *)findFreeBlock_noLock(s, d, e);	// space for entry
	
	setBlockDirty(b);
		
	e->block_ptr = b;						 // set block_ptr	
						
	if( S_ISDIR(e->a.st_mode) )  // check dir mode entry
	{	
		uint64_t block_nr = getFreeBlock_noLock(s);	// get block for dir
		
		e->data = block_nr;				 // link entry with dir			
				
		if( !createDir_noLock(block_nr) )	{
			perror("Problem createDir");
			return false;
		}			
	}			
 	if( S_ISREG(e->a.st_mode) ) 	//  create file mode entry
	{		
		e->data = 0;		// 0 -> no data add TODO future work
	}	
	if( !appendEntry(b,e) )	{
		fprintf(stderr,"%s \n", "error appendEntry" );
		return false;
	}

	g_tree_insert(d->t, e->name, e); // insert in tree with entries
			
	WRITE_UNLOCK
			
	return true;
}

bool mefs_chmod(Entry * e, mode_t mode)
{	
	WRITE_LOCK	
	e->a.st_mode = mode;			
	setBlockDirty(e->block_ptr);	
	WRITE_UNLOCK	
	return true;
}

bool mefs_chown(Entry * e, uid_t uid, gid_t gid)
{	
	WRITE_LOCK	
	e->a.st_uid = uid;
	e->a.st_gid = gid;	
	setBlockDirty(e->block_ptr);	
	WRITE_UNLOCK	
	return true;
}

bool mefs_utimens(Entry * e, struct timeval ts[2])
{
	WRITE_LOCK
	e->a.st_atime = ts[0].tv_sec;
	e->a.st_mtime = ts[1].tv_sec;	
	setBlockDirty(e->block_ptr);	
	WRITE_UNLOCK	
	return true;
}

bool mefs_rename(Super * s, Entry * e)
{		
	Block * b = e->block_ptr;	// get parent block
		
	WRITE_LOCK
	
	b->bs->fillpos -= e->name_len + CONST_PHYSICAL_ENTRY_SIZE; // calc new fillpos		
		
	removeEntry(b->dir_ptr, e);	
	
	free(e);
	
	setBlockDirty(b);
	
	WRITE_UNLOCK
	
	return true;
	
}

bool appendEntry(Block * b, Entry * e)
{	
	if(dbg > 0) {printf("<appendEntry>"); displayEntry(e);}
	
	if( b->e == NULL)
	{
		b->e = b->last_e = e;			
	}
	else
	{ 
		e->prev = b->last_e;  
		e->next = NULL;
		b->last_e->next = e;  
		b->last_e = e;		
	}	
  b->bs->fillpos += e->name_len + CONST_PHYSICAL_ENTRY_SIZE;
				
	e->block_ptr = b;	  
	
	return true;
}

bool compressBlocks(Super * s, Dir * d)
{
  // not used just for memory TODO
  
	Block * b1 = d->b->next;	// skip first block
	Block * b2 = b1->next;		// init compare block
	
	uint32_t cur_fillpos = 0;
	
	while(b1 != NULL)
	{
		cur_fillpos = b1->bs->fillpos;
		while(b2 != NULL)
		{	
			if(cur_fillpos + b2->bs->fillpos < s->block_size)
			{
				b1->last_e->next = b2->e;		// append old entry list 
				b1->last_e = b2->last_e;		// set new end of list
				b1->bs->fillpos += b2->bs->fillpos;		// calc new fillpos
				b2->prev->bs->ext_block_nr = b2->bs->ext_block_nr;	// chnage ext_block_nr
				removeBlock(b2); break;
			}
			b2 = b2->next;
		}
		b1 = b1->next;
	}	
	return true;
}

bool isDirEmpty(Dir * d)
{		
	Block * b = d->b;	
	while(b != NULL)
	{		
		if (b->bs->fillpos != 0) return false; 	
		b = b->next;	  
	} 	
	return true;
}

bool removeDir(Super * s, uint64_t dir_nr)
{
	if(dbg > 0)  printf("<removeDir>  %"PRIu64" \n", dir_nr);		
		
	Dir * d = (Dir *)mefs_lookupDir(dir_nr); // search in loaded Dirs	
		
	if(d == NULL) 	// Dir not oloaded => load dir
	{	
		d = mefs_loadDir(dir_nr);			
	}		
		
	if ( g_tree_nnodes (d->t) != 0) return false;
		
	Block * b = d->b;	
	while(b != NULL)
	{		
		setBlockFree_noLock(s, b->block_nr);
		//unloadBlock(b);
		b = b->next;	  
	} 	
					
	g_tree_steal( fs.dt , &d->dir_nr );	// remove from loaded dirs
	
	g_tree_destroy(d->t);	
	
	free(d);
		
	return true;		
		
}	


bool mefs_delete(Super * s, Entry * e)
{	
	/*
	if( S_ISDIR(e->a.st_mode) ) 
	{			
		if ( !removeDir(s, e->data) )
		return false;
	}
	*/	
	Block * b = e->block_ptr;	// get parent block
		
	WRITE_LOCK
	
	b->bs->fillpos -= (e->name_len + CONST_PHYSICAL_ENTRY_SIZE ); // calc new fillpos	
			
	if (b->bs->fillpos == 0 && b->bs->ext_block_nr) 	
	setBlockFree_noLock(s, b->block_nr);	
	
	removeEntry(b->dir_ptr, e);		
	
	free(e);
	
	setBlockDirty(b);
	
	WRITE_UNLOCK
	
	return true;
}	
	
bool removeEntry(Dir * d, Entry * e)
{					
	if(dbg > 0){ printf("<removeEntry> "); displayEntry(e); }
	
	g_tree_steal( d->t , e->name );		// remove from tree	
	
	Block * b = e->block_ptr;	
	
	if ( (e->prev == NULL) && (e->next == NULL) ) // only one entry?
	{ 
		b->e = NULL;
	}	
  else if (e->prev == NULL)   // first entry?
  { 	
    b->e = e->next;
    e->next->prev = NULL;
  }
  else if (e->next == NULL) 	// last entry?
  { 
    b->last_e = e->prev;
    e->prev->next = NULL;
  } 		
  else 
  {  
    e->prev->next = e->next;
    e->next->prev = e->prev;
	}	
		
	return true;
}

#define COPY_BUFFER(src, tgt, offset, size) \
  memcpy(tgt, src, size); \
  offset += size;
  
bool statBlockToBuff(Block * b, char * buff, off_t * curBuffPos)
{
#define WRITE_L_BUFFER(src, size) COPY_BUFFER(src, buff + *curBuffPos, *curBuffPos, size)

  WRITE_L_BUFFER(& b->bs->fillpos, sizeof(uint32_t) )
  WRITE_L_BUFFER(& b->bs->ext_block_nr, sizeof(uint64_t) ) 
	
#undef WRITE_L_BUFFER
	
	return true;
}

  	
bool entryToBuff(Entry * e, char * buff, off_t * curBuffPos)
{	
#define WRITE_L_BUFFER(src, size) COPY_BUFFER(src, buff + *curBuffPos, *curBuffPos, size)
 
  WRITE_L_BUFFER(& e->name_len, sizeof(uint8_t))
	WRITE_L_BUFFER(& e->a.st_mode, sizeof(uint32_t))
	WRITE_L_BUFFER(& e->a.st_uid, sizeof(uint32_t))
	WRITE_L_BUFFER(& e->a.st_gid, sizeof(uint32_t))	
	WRITE_L_BUFFER(& e->a.st_ctime, sizeof(uint64_t))
	WRITE_L_BUFFER(& e->a.st_atime, sizeof(uint64_t))
	WRITE_L_BUFFER(& e->a.st_mtime, sizeof(uint64_t))
	WRITE_L_BUFFER(& e->data, sizeof(uint64_t))	
  WRITE_L_BUFFER(e->name, e->name_len)
	
#undef WRITE_L_BUFFER
	
	return true;
}

gboolean iter_all_loaded_dirs( gpointer nr, Dir * d)
{  	
	unloadDir(d);	
  return FALSE;
}

gboolean iter_all_loaded_dirs_find_small( gpointer nr, Dir * d)
{  		
	if(d->fillpos < 200  ) 
		{ unloadDir(d); return FALSE;}	
  return TRUE;
}

bool unloadLoadedDirs()
{
	puts("--------------------------------------");
	printf("<unloadLoadedDirs>  Tree nNodes %d \n", g_tree_nnodes(fs.dt) );
	puts("--------------------------------------");
	
	g_tree_foreach(fs.dt, (GTraverseFunc)iter_all_loaded_dirs, NULL);
	
	return true;
}

bool unloadLoadedDirs_small()
{
	puts("--------------------------------------");
	printf("<unloadLoadedDirs_small>  Tree nNodes %d \n", g_tree_nnodes(fs.dt) );
	puts("--------------------------------------");
	
	g_tree_foreach(fs.dt, (GTraverseFunc)iter_all_loaded_dirs_find_small, NULL);
	
	return true;
}


Block * initBlock(uint64_t block_nr)
{
  Block * b = (Block *) calloc(1, sizeof(Block));
		
	if( b == NULL )
	{
		fprintf(stderr,"%s \n", "error initBlock" );
		return NULL;
	}	
	b->block_nr = block_nr;
		
	b->bs = (Blockstat *) calloc(1, sizeof(Blockstat) ); 	
				   
  //printf("<initBlock> nr %"PRIu64" - %p \n",	block_nr , b);
   
  return b;
}

Blockstat * readBlockStat(Super * s, uint64_t block_nr)
{
	Blockstat * bs = (Blockstat *) calloc(1, sizeof(Blockstat) ); 
			     	     
  ssize_t ret = pread(s->file,	bs,	CONST_BLOCK_HEADER_SIZE,	block_nr * s->block_size);
     	
	if(ret < CONST_BLOCK_HEADER_SIZE)
	{
		fprintf(stderr, "Problem readBlockStat read %llu of %d Bytes \n", (long long unsigned) ret, CONST_BLOCK_HEADER_SIZE);
		return false;
	}
  		
  return bs;
}

bool readBlockEntries(Block * b, char * buff)
{    		
	off_t offset =  fs.s->block_size * b->block_nr + CONST_BLOCK_HEADER_SIZE;
    		
  ssize_t ret = pread(fs.s->file,	buff,	b->bs->fillpos, offset);
	if(ret < b->bs->fillpos){
		perror("Problem readBlockEntries");
    return false;
	} 
	printf("<readBlockEntries> block_nr %"PRIu64" at %"PRId64" fillpos %"PRIu32"\n", b->block_nr , offset, b->bs->fillpos);
  
  return true;
}

Block * loadBlock(uint64_t block_nr)
{
  Block * b = (Block *) initBlock(block_nr);
		
	b->bs = readBlockStat(fs.s, block_nr); 
	
	//fs.s->fillpos +=  b->bs->fillpos;		// count fill positions of loaded blocks
	
	//printf("<loadBlock>  fs.s.fillpos: %"PRId64" ", fs.s->fillpos ); displayBlockStat(b);		//dbg
		 
	ssize_t buffSize = b->bs->fillpos;   
			  
	char * buff = (char *) calloc(1, buffSize + 1 );
		
  readBlockEntries(b, buff);   // read block entries from disk to buffer
	
 	b->bs->fillpos = 0;		 // reset to calculate new fillpos by append -> check
			
	off_t curBuffPos = 0;
	
	Entry * e;
	
  while ( curBuffPos < buffSize )
  {
    e = (Entry *) calloc(1, sizeof(Entry)); 	
		     
    buffToEntry( e, buff, &curBuffPos);		// read entry form buffer
														    
		appendEntry(b, e);					
  } 
					
	free(buff);	
	
	/*
	fs.loaded_block_count += 1;
	
	if(fs.loaded_block_count >=fs.max_cached_block_count )
	{
		if ( (unloadLoadedDirs_small()) ) return b; return NULL;
	} 
	*/  
  return b;
}

static bool writeBackBlock(Block * b, int file, off_t offset, bool write_full_block)
{
	assert(CONST_BLOCK_HEADER_SIZE + b->bs->fillpos <= fs.s->block_size);
	
	// for write performance check !
	ssize_t buffSize = write_full_block ? (fs.s->block_size) :  (CONST_BLOCK_HEADER_SIZE + b->bs->fillpos);   
	
	char * buff = (char *)calloc(1, buffSize);
			
	off_t curBuffPos = 0;
   	
  statBlockToBuff(b, buff, &curBuffPos);
	
	Entry * e = b->e;
	
  while( e != NULL )
  {		
		entryToBuff(e, buff, &curBuffPos);
		e = e->next;		
  }		
	
	ssize_t ret = pwrite(file, buff, buffSize, offset );
    		
	if ( ret < buffSize )	{
	  fprintf(stderr, "Problem writeBlock - Bytes written:  %llu \n", (long long unsigned) ret);
	}
	printf("done <writeBlock> block_nr %"PRIu64" Bytes %llu\n", b->block_nr, (long long unsigned) ret);
	
	free(buff);
		
	printf("<writeBackBlock> "); //displayBlockStat(b);		//dbg
	
	return true;
}

 bool unloadBlock(Block * b)
{  	
	if (dbg > 0){printf("<unloadBlock>  "); displayBlockStat(b); }		
		
	Entry * e = b->e, *cur;
	
	Dir * d = b->dir_ptr;
	
  while( e != NULL )
  {
		cur = e;
		removeEntry(d, cur);			
    e = e->next;		
		free(cur);
  }					
	removeBlock(b);						
	
  return true;
}

 bool removeBlock(Block * b)
{	
	Dir * d = b->dir_ptr;
	
	if (b->prev == NULL && b->next == NULL)    		// one block 
	{ puts("--removeBlock first and last => one ");
		d->b = NULL;
		d->last_b = NULL;
	}
	else if (b->prev == NULL && b->next != NULL)	//first block
	{ puts("--removeBlock first");
		d->b = b->next;
		b->prev = NULL;
	}
	else if (b->prev != NULL && b->next == NULL)   // last block
	{ puts("--removeBlock last");
		d->last_b = b->prev;
		b->prev->next = NULL;
	} 		
	else 
	{ puts("--removeBlock in between ");
		b->prev->next = b->next;
		b->next->prev = b->prev;
	}	
	printf("<removeBlock> nr %"PRIu64" - %p  \n",b->block_nr, b);	
	
	free(b->bs);
	free(b);
	
	return true;
}

 bool appendBlock(Dir * d, Block * b)
{
  b->dir_ptr = d;
	
	if( d->b == NULL)
  {	
    d->b = d->last_b = b;
	}
  else
  {
    b->prev = d->last_b;
		d->last_b->next = b;
		d->last_b = b;
  }				 
  printf("<appendBlock> block_nr %"PRIu64" - %p  to dir_nr %"PRIu64" - %p\n",
												b->block_nr, b, d->dir_nr, b->dir_ptr);	
  return true;
}

extern bool buffToEntry(Entry * e, char * buff, off_t * curBuffPos)
{
#define READ_L_BUFFER(tgt, size) COPY_BUFFER(buff + *curBuffPos, tgt, *curBuffPos, size)
 
  READ_L_BUFFER(& e->name_len,   sizeof(uint8_t))
  READ_L_BUFFER(& e->a.st_mode,  sizeof(uint32_t))
	READ_L_BUFFER(& e->a.st_uid,   sizeof(uint32_t))
	READ_L_BUFFER(& e->a.st_gid,   sizeof(uint32_t))	
	READ_L_BUFFER(& e->a.st_ctime, sizeof(uint64_t))
	READ_L_BUFFER(& e->a.st_atime, sizeof(uint64_t))
	READ_L_BUFFER(& e->a.st_mtime, sizeof(uint64_t))	
	READ_L_BUFFER(& e->data,       sizeof(uint64_t))	
  
  e->name = calloc(1, e->name_len + 1);
  
  READ_L_BUFFER(e->name, e->name_len)
  
#undef READ_L_BUFFER

return true;
}

extern bool initBitmap(Super * s)
{
	s->bitmap = (uint32_t *) calloc(1, s->bitmap_bsize);
				
	if(s->bitmap == NULL) {
      fprintf(stderr, "Problem initBitmap - allocation \n");
      return false;
  }	
  
  uint32_t i;			// set bitmap 
  for(i = 0; i < s->bitmap_size; i++)
	{
		s->bitmap[i] = 0;
  }
	
	return true;
}

extern bool writeBitmap(Super * s)
{				
	ssize_t ret = pwrite( s->file, s->bitmap, s->bitmap_bsize, s->block_size );
		
	if(ret < s->bitmap_size * 4 ) {
      fprintf(stderr, "Problem writeBitmap - written Bytes %llu of %d \n", (long long unsigned) ret, s->bitmap_size * 4 );
      return false;
  }		
	printf("<writeBitmap> written Bytes  %i  of %d \n", ret, s->bitmap_size * 4); 
  
  return true;
}

extern bool readBitmap(Super * s)
{
	off_t offset = s->block_size;  // seek super block read at block 1
				
	ssize_t ret = pread( s->file, s->bitmap, s->bitmap_bsize, offset );
			  	
	if(ret < s->bitmap_bsize ) {
      fprintf(stderr, "Problem readBitmap - read Bytes %llu of %d \n", (long long unsigned) ret, s->bitmap_bsize );
      return false;
  }	
	printf("<readBitmap> read Bytes %"PRIu32" of %"PRIu32" @ %"PRIu32"\n",(unsigned) ret, s->bitmap_bsize, (unsigned)offset); 
  
  return true;
}

extern Super * initSuper()
{	
	Super * s = (Super *) calloc(1, sizeof(Super));	
		
	s->root = (Entry *) calloc(1, sizeof(Entry));
			
	if(s->root == NULL) {
      fprintf(stderr, "Problem initSuper - Root allocation \n");
      return false;
  }		
	return s;  
}

extern bool calculateSuper(Super * s)
{
	s->blocks = (uint64_t)(s->volume_size / s->block_size);	
	assert(s->blocks > 0);
	s->blocks = (	s->blocks % 2) ? (s->blocks - 1) : s->blocks;		
		
	s->bitmap_size = (uint32_t)(s->blocks / 32 + 1);	// bitmap array component count 
	s->bitmap_bsize = s->bitmap_size * 4;	
		
	return true;
}

extern bool createSuper(Super * s)		/* only for create_metafs */
{				
	setBlockUsed_noLock(s,0);		// Super BlocK
	setBlockUsed_noLock(s,1);		// Bitmap Block
		
	s->root = mefs_initEntry("/", (S_IFDIR | 0777) );
	
	s->root->data = 2;	
	setBlockUsed_noLock(s, s->root->data);		
	
	return true;
}

extern bool superToBuff(Super * s, char * buff, off_t * curBuffPos)
{			
#define WRITE_L_BUFFER(src, size) COPY_BUFFER(src, buff + *curBuffPos, *curBuffPos, size)

  WRITE_L_BUFFER(& s->volume_size, sizeof(uint64_t) )
  WRITE_L_BUFFER(& s->block_size, sizeof(uint32_t) ) 
	WRITE_L_BUFFER(& s->dirty, sizeof(uint8_t) ) 
	
#undef WRITE_L_BUFFER

	return true;
}

extern bool buffToSuper(Super * s, char * buff, off_t * curBuffPos)
{			
#define READ_L_BUFFER(tgt, size) COPY_BUFFER(buff + *curBuffPos, tgt, *curBuffPos, size)
  	
  READ_L_BUFFER(& s->volume_size, sizeof(uint64_t) )
  READ_L_BUFFER(& s->block_size, sizeof(uint32_t) ) 
	READ_L_BUFFER(& s->dirty, sizeof(uint8_t) ) 
    
#undef READ_L_BUFFER
	
	return true;
}

extern bool readSuper(Super * s)
{	
	ssize_t buffSize =  13 + CONST_PHYSICAL_ENTRY_SIZE + 1  ;	// const + "/"
	
	char * buff = calloc(1, buffSize + 1 );
	
	//ssize_t ret = pread(s->file, buff, buffSize, 0);
			
  ssize_t ret = pread(s->file, buff, buffSize, 0);	// read Super Block @ block 0
		
	if(ret < buffSize ) {
      fprintf(stderr, "Problem readSuper - read Bytes %llu of %llu \n", (long long unsigned) ret,  (long long unsigned) buffSize );
      return false;
  }
	
	off_t curBuffPos = 0;
	
	buffToSuper( s, buff, &curBuffPos );
	
	buffToEntry( s->root, buff, &curBuffPos );
			
	free(buff);
  return true;
}

extern bool writeSuper(Super * s)
{		
	ssize_t buffSize = 13 + CONST_PHYSICAL_ENTRY_SIZE + 1  ;		// const + "/"
	
	char * buff = calloc(1, buffSize + 1 );	
	
	off_t curBuffPos = 0;	
	
	superToBuff(s, buff, &curBuffPos);
	
	entryToBuff(s->root, buff, &curBuffPos);
	
	puts("<writeSuper>");	  
			
	ssize_t ret = pwrite(s->file, buff, buffSize, 0);	// write Super Block @ Block 0
		 	
	if(ret < buffSize ) {
      return false;
  }	
	free(buff);		
	
  return true;
}

 bool initLoadedDirsTree()
{
	fs.dt = g_tree_new((GCompareFunc)block_nr_cmp);
	
	if(fs.dt == NULL) {
      fprintf(stderr, "Problem initSuperTree - allocation \n");
      return false;
  }
	return true;
}


extern bool setBlockUsed_noLock(Super * s, uint64_t block_nr)
{
	uint32_t byte_pos = (uint32_t)(block_nr / 32);   
			
	uint32_t bit_pos = block_nr % 32; 	 
	
	uint32_t bitmask = (uint32_t)1 << bit_pos; // set bitmask
	
	assert(byte_pos >= 0);
	assert(byte_pos < s->bitmap_size);
	
	s->bitmap[byte_pos] = s->bitmap[byte_pos] | bitmask;
	
	//setBlockDirty(2); 	// static bitmap write back

	return true;
}  

extern bool setBlockFree_noLock(Super * s, uint64_t block_nr)
{
	uint32_t byte_pos = (uint32_t)(block_nr / 32);	// get array position
		
	uint32_t bit_pos = block_nr % 32;		// get bit position 
		
	uint32_t bitmask = (uint32_t)1 << bit_pos; // set bitmask
	
	assert(byte_pos >= 0);
	assert(byte_pos < s->bitmap_size);
	
	s->bitmap[byte_pos] = s->bitmap[byte_pos] ^ bitmask;
	
	//setBlockDirty(2);		// static bitmap write back each x sec

	return true;
}

extern  uint64_t getFreeBlock_noLock(Super * s)
{
	uint64_t i, j, block_nr;
	
	for( i=0; i < s->bitmap_size; i++)	// search byte_pos in array
	{
		if(  s->bitmap[i] < 0xFFFFFFFF ) break;	// search block group 
	}
	
	uint64_t tmp = i * 32LLU;		// set tmp block_nr to byte_pos 
		
	for( j=0; j<32; j++)		// find free Bit = 0
	{
		if ( (s->bitmap[i] >> j) % 2 == 0)
		{			
			block_nr = tmp + j;
			break;
		}
	}		
	setBlockUsed_noLock(s, block_nr);		
	
	return block_nr;
}

extern uint64_t countFreeBlocks(Super * s)
{
	uint64_t i, j, c=0;		
	for( i=0; i < s->bitmap_size-1 ; i++)
	{
		for( j=0; j<32; j++)
		{
			if ( ( (s->bitmap[i] >> j) % 2) == 0)
				c++;			
		}
	}
	return c;
}	

bool displayBitmap(Super * s)
{
	puts("------------------------------------------------------------");
	printf("<dispaly Bitmap> size: %"PRIu32" Byte | free blocks: %"PRIu64" \n", s->bitmap_bsize, (uint64_t)countFreeBlocks(s) );
	puts("------------------------------------------------------------");
		
	uint32_t j, i;
			
	for( i=0; i < s->bitmap_size; i+=2)		// iterate through bitmap
	{		
		for( j=0; j<32; j++ )
		{
			if ( (s->bitmap[i] >> j) % 2 == 0)		// display part of bitmap - two units in one line
				printf("0");
			else	printf("1");
		}
		for( j=0; j<32; j++ )
		{
			if ( (s->bitmap[i+1] >> j) % 2 == 0)		// display part of bitmap
				printf("0");
			else	printf("1");
		}
		printf(" : Bitmap[%d-%d] \n", i, i+1); //printf("%"PRIu32" \n ", s->bitmap[i]);
		
	}	
	return true;
}

bool loadTree(Dir * d)
{	
	//printf("<loadTree> \n ");
  Block * b = d->b;
	    
  while(b != NULL)
  { 
		Entry * e = b->e;
		
    while(e != NULL)
    {      
			//displayEntry(e);	//dbg
																	
			g_tree_insert(d->t, e->name, e);
		
			e = e->next;
    } 		
		b = b->next;	
  }
  return true;
}
																												
Dir * initDir(uint64_t block_nr)
{	
  Dir * d = (Dir *) calloc(1,sizeof(Dir));

	if (d == NULL)  {
    fprintf(stderr, "Problem initDir \n");
    return NULL;
  }	
	d->dir_nr = block_nr;	
		  
  d->t = g_tree_new((GCompareFunc)g_ascii_strcasecmp);
	
  if (d->t == NULL)  {
    fprintf(stderr,"Problem initDir Tree\n");
    return NULL;
  }
	printf("<initDir> dir_nr %"PRIu64" - %p	\n", d->dir_nr, d);
	return d;
}

extern Dir * mefs_loadDir(uint64_t block_nr)
{
	printf("<mefs_loadDir> block_nr %" PRIu64 "\n", block_nr);  
	Dir * d = (Dir *) initDir(block_nr);
	
	if( d == NULL)
		return NULL;		
					 
  Block * b = (Block *) loadBlock(block_nr);  
  if( b == NULL)
		return NULL;		
		
	//b->dir_ptr = d; 
  
  appendBlock(d,b);
  
  while(b->bs->ext_block_nr != 0)
  {    	
		b = loadBlock(b->bs->ext_block_nr);
	
		//b->dir_ptr = d;
       
    appendBlock(d,b);
	}		
  loadTree(d);
	
	appendDir(d);
  	
  return d;
}
 
 bool unloadDir(Dir * d)
{
	printf("<unloadDir>  %" PRIu64 "\n", d->dir_nr);
	
	Block * b = d->b;	
			
	while (b != NULL)		// iter block list
	{		
		unloadBlock(b);		// free all entries	and remove block from list
		b = b->next;		
	}	
	g_tree_destroy(d->t);
		
	g_tree_steal( fs.dt , &d->dir_nr );		// remove from loadedDirsTree 
		
	free(d);
			
	return true;
}

 gint block_nr_cmp(const guint64 * a, const guint64 * b)
{
	if( *a < *b)  return -1; 
		 if( *a == *b) return 0; 
			 return 1; 	// ( *a > *b)
}
  	
 bool appendDir( Dir * d)
{
	printf("<appendDir> dir_nr %"PRIu64"- %p \n", d->dir_nr, d);
	
	g_tree_insert(fs.dt, (gpointer)&d->dir_nr, (gpointer) d);
	
	return true;
}
	
extern Dir * mefs_lookupDir(uint64_t dir_nr)
{		
	Dir * d = NULL;
	
  uint64_t * ok = NULL;
	
  gconstpointer key = (gpointer *)&dir_nr;
    
	gboolean found = FALSE;
		 
  found = g_tree_lookup_extended (fs.dt, key, (gpointer*)&ok, (gpointer*)&d );
		
	if(!found) return NULL;
		
  //printf("<lookupDir> found key %"PRIu64" val: %p \n",(uint64_t)*ok , (Dir *)d);   
	
  return d;
}

extern Entry * mefs_lookupEntry(Dir * d, char * name)
{
	Entry * e = NULL;
	
  gpointer * ok = NULL;
	
  gconstpointer key = (gpointer *)name;
  
	gboolean found = FALSE;
	
  found = g_tree_lookup_extended (d->t, key, (gpointer *)&ok, (gpointer *)&e );
	
	if(!found) return NULL;
	
	//printf("<lookupEntry> found key %s val: %p  data %"PRIu64"\n", (char *) ok, e, e->data);  
	
	return e;
}
 
extern GList * buildLookupList(const char * path)
{
	GList * lookup_list = NULL;
	
	char * cur_name;	
	
	while ( strcmp(path , "/") )
	{				
		cur_name = g_path_get_basename(path);
					
		lookup_list = g_list_append(lookup_list, cur_name);
								
		path = g_path_get_dirname(path);
	}		
	return lookup_list;
}

extern Entry * mefs_lookupPath(const char * path) 
{
	GList * ll = buildLookupList(path);
		
	guint c = g_list_length(ll);			
	
	if( c == 0) return fs.s->root;	// path empty => root dir => already loaded
  
  Dir * d = fs.d;	  // the root-dir of the filesystem
	
	char  * cur_name = (char *)g_list_last(ll)->data; 	
						
	Entry * e = (Entry *)mefs_lookupEntry(d,  cur_name);	
						
	//ll = g_list_remove(ll, cur_name);  c -= 1;	
	ll = g_list_delete_link( ll, g_list_last(ll) );  c -= 1;	
			
	while(c > 0 && e != NULL && S_ISDIR(e->a.st_mode)  ) 
	{
		printf("<lookupPath>  %s - %p data -> %"PRIu64"\n", e->name, e, e->data);
		
		d = (Dir *) mefs_lookupDir(e->data);			
		
		if( d == NULL )		// if not loaded but exist
		{
			d = mefs_loadDir(e->data);
				
			if( d == NULL) return NULL;			// error			
		}
		cur_name = g_list_last(ll)->data;
		
		e = (Entry *)mefs_lookupEntry(d, cur_name);	
		
		if(e == NULL)		// entry not found
		{
			printf("<lookupPath> ENOENT \n");
			g_list_free(ll);
			return NULL;
		}			
		//ll = g_list_remove(ll, cur_name);		--c;	
		ll = g_list_delete_link( ll, g_list_last(ll) );	 --c;	
	}
	g_list_free(ll);	
		
	return e;
}
	
extern bool displaySuper(Super * s)
{	
	s->free_blocks = (uint64_t)countFreeBlocks(s);
	
	s->used_blocks = s->blocks - s->free_blocks;
	
	char status[6];
		
	s->dirty  ?  strcpy(status, "dirty") : strcpy(status, "clean");		
	
	puts("---------------------------------------------------------------");
	printf("<displaySuper>  File System Parameter: status = %s \n", status);  
	
	printf("volume_size: %"PRIu64" Byte = %"PRIu64" MB\n", s->volume_size, s->volume_size/(1024*1024) );
	
	printf("block_size : %"PRIu32" Byte = %"PRIu32" KB\n", s->block_size, s->block_size/1024 );
		
	printf("blocks : %"PRIu64" | free : %"PRIu64" | used : %"PRIu64" \n", s->blocks, s->free_blocks, s->used_blocks);
		
	printf("bitmap_size: %"PRIu32" Units = %"PRIu32" Byte \n", s->bitmap_size , s->bitmap_bsize );
		
	displayEntry(s->root);	
	
	return true;
}	


bool mefs_destroy()
{
  printf("Destroy\n");
  
  WRITE_LOCK
  fs.closed = true;
  pthread_cond_signal(& fs.enforce_write_back_cond);  
  WRITE_UNLOCK
    
  if (pthread_join(fs.write_back_thread, NULL) != 0)
  {
     return EXIT_FAILURE;
  }     	
	//displayBitmap(fs.s);
	writeBitmap(fs.s);
	
	fs.s->dirty = 0;
	
	writeSuper(fs.s);	

  close(fs.s->file);
  close(fs.s->data_file);
	
	return true;
}

extern Filesystem mefs_loadFilesystem(char * metadataFile, char * dataFile)
{  	
	signal(SIGINT, & signalHandler);
	
	pthread_mutex_init (& fs.write_mutex, NULL);
	pthread_cond_init (& fs.enforce_write_back_cond, NULL);
	
  fs.max_cached_block_count  = 1000;    // TODO
  
  fs.max_cached_byte_count  =  200*1024*1024;		// static set cache size
		
	fs.closed = false;						// set FS active  
	
	fs.dirty_block_list = NULL;		// init dirty block list
  
	fs.s =(Super *) initSuper(); 
	
	fs.s->file  = open(metadataFile, O_RDWR );	// Meta Data storage
	
	fs.s->data_file  = open(dataFile, O_RDWR );	// Log File aka Data
	
	readSuper(fs.s);		
		
	calculateSuper(fs.s);		// calculate further FS parameter 
			
	initBitmap(fs.s);
				
	readBitmap(fs.s);
	
	//displaySuper(fs.s);	// dbg
	
	displayBitmap(fs.s);	// dbg
	
	if(fs.s->dirty == 1) 
	{
		fprintf(stderr, "fs not clean - check ! \n");
		//return false;
	}	
	fs.s->dirty = 1;		// set dirty flag => fs is in use
	
	writeSuper(fs.s);		// write dirty flag to disk => update Super Block
	
	initLoadedDirsTree(fs.dt);			// loaded Directory Cache top structure 
		
	fs.d =(Dir *) mefs_loadDir(fs.s->root->data);	// load Root-Dir content - when change => change path_lookup()
					
  if (pthread_create(& fs.write_back_thread, NULL, write_back_thread_function, NULL) != 0)
  {
    exit(1);
  }
  
	return fs;
}

	
